#include <iostream>

#include "yaml.h"
#include "pdg.h"
#include "clast_scanner.h"
extern "C" {
#include "piputil.h"
}
#include "cloog_util.h"

using namespace std;
using namespace pdg;

#define MAXRAYS    POL_NO_DUAL

struct core_printer : public clast_scanner {
    void print_user(clast_user_stmt *s);
};

void core_printer::print_user(clast_user_stmt *u)
{
    pdg::dependence *dep = (pdg::dependence*) u->statement->usr;
    pdg::node* from = dep->from;
    pdg::node* to = dep->to;
    int from_access = dep->from_access ? dep->from_access->nr : -1;
    int to_access = dep->to_access ? dep->to_access->nr : -1;

    print_indent();
    if (u->statement->number == 1) {
	fprintf(dst, "if (++i_%d_%d_%d_%d > s_%d_%d_%d_%d)\n",
		    from->nr, from_access, to->nr, to_access,
		    from->nr, from_access, to->nr, to_access);
	print_indent();
	fprintf(dst, "    s_%d_%d_%d_%d = i_%d_%d_%d_%d;\n",
		    from->nr, from_access, to->nr, to_access,
		    from->nr, from_access, to->nr, to_access);
    } else
	fprintf(dst, "--i_%d_%d_%d_%d;\n",
		    from->nr, from_access, to->nr, to_access);
}

struct program_printer : public clast_scanner {
    PDG *pdg;

    void print_user(clast_user_stmt *s);
    void scan_program(CloogProgram * p, CloogOptions *options);
    void print_program(PDG *pdg);
};

void program_printer::print_user(clast_user_stmt *s)
{
    CloogOptions *options = cloog_options_malloc();
    options->compilable = 1;
    options->l = pdg->dimension;

    int nparam = pdg->params.size();
    Polyhedron *ctx = pdg->context ? (Polyhedron *) *pdg->context : 
				     Universe_Polyhedron(nparam);

    cloog_program_builder pb(ctx);

    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type == pdg::dependence::uninitialized)
	    continue;
	pdg::node* from = dep->from;
	pdg::node* to = dep->to;
	int from_access = dep->from_access ? dep->from_access->nr : -1;
	int to_access = dep->to_access ? dep->to_access->nr : -1;
	if (!dep->size)
	    continue;
	print_indent();
	fprintf(dst, "int i_%d_%d_%d_%d = 0;\n", from->nr, from_access,
						 to->nr, to_access);
	print_indent();
	fprintf(dst, "int s_%d_%d_%d_%d = 0;\n", from->nr, from_access,
						 to->nr, to_access);
	pdg::UnionMap *rel = scatter_dependence(dep);
	Polyhedron *D = *rel;
	for (Polyhedron *P = D; P; P = P->next) {
	    Polyhedron *read = Polyhedron_projectout_dep(P, rel->input);
	    pb.add_domain(read, -1, dep);
	}
	for (Polyhedron *P = D; P; P = P->next) {
	    pb.add_domain(Polyhedron_Copy(P), 1, dep);
	}
	rel->params.resize(0);
	rel->free();
	delete rel;
	Domain_Free(D);
    }

    pb.set_names(pdg->params, 0);

    pb.p = cloog_program_generate(pb.p, options);

    clast_stmt *stmt = cloog_util_create_clast(pb.p, options);
    core_printer cp;
    cp.indent = indent;
    cp.dst = dst;
    cp.print_iterators(pb.p->names);
    cp.print(stmt);
    cloog_clast_free(stmt);

    cloog_program_free(pb.p);
    cloog_options_free(options);

    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type == pdg::dependence::uninitialized)
	    continue;
	pdg::node* from = dep->from;
	pdg::node* to = dep->to;
	int from_access = dep->from_access ? dep->from_access->nr : -1;
	int to_access = dep->to_access ? dep->to_access->nr : -1;
	if (!dep->size)
	    continue;
	print_indent();
	fprintf(dst, "assert(i_%d_%d_%d_%d == 0);\n",
			    from->nr, from_access, to->nr, to_access);
	print_indent();
	fprintf(dst, 
	    "printf(\"s_%d_%d_%d_%d: %%d; %%d\\n\", s_%d_%d_%d_%d, (int)(%s));\n", 
			    from->nr, from_access, to->nr, to_access,
			    from->nr, from_access, to->nr, to_access,
			    dep->size->s->c_str());
	print_indent();
	fprintf(dst, "assert(s_%d_%d_%d_%d <= (int)(%s));\n",
			    from->nr, from_access, to->nr, to_access,
			    dep->size->s->c_str());
    }
}

void program_printer::scan_program(CloogProgram * p, CloogOptions *options)
{
    clast_stmt *stmt = cloog_util_create_clast(p, options);

    print_header();
    fprintf(dst, "%s", "#define max(x,y)\t" "((x)>(y)?(x):(y))" "\n");
    fprintf(dst, "\n");
    fprintf(dst, "int main()\n{\n");
    indent += 4;
    print_iterators(p->names);
    print(stmt);
    print_indent();
    fprintf(dst, "return 0;\n");
    fprintf(dst, "}\n");
    cloog_clast_free(stmt);
}


void program_printer::print_program(PDG *pdg)
{
    dst = stdout;
    this->pdg = pdg;

    Polyhedron *ctx = (Polyhedron *) *pdg->context;

    CloogOptions *options = cloog_options_malloc();
    options->compilable = 1;

    cloog_program_builder pb;

    pb.p->nb_scattdims = 0;
    pb.p->scaldims = NULL;

    pb.add_domain(ctx, 0);

    pb.p->names->nb_parameters = 0;
    pb.p->names->parameters = NULL;
    pb.p->names->nb_iterators = ctx->Dimension;
    pb.p->names->iterators = cloog_util_parameter_names(pdg->params);
    pb.p->names->scattering = NULL;
    pb.p->names->nb_scattering = 0;
    pb.p->names->nb_scalars = 0;

    pb.p = cloog_program_generate(pb.p, options);
    scan_program(pb.p, options);
    cloog_program_free(pb.p);
    cloog_options_free(options);
}

int main(int argc, char * argv[])
{
    PDG *pdg;
    pdg = yaml::Load<PDG>(stdin);

    assert(pdg->context);

    program_printer pp;
    pp.print_program(pdg);

    pdg->free();
    delete pdg;

    return 0;
}
