#include <stdio.h>
#include <string.h>
#include <iostream>
#include <map>
#include "argp.h"
#include "progname.h"

#include "version.h"

#include "yaml.h"
#include "pdg.h"
#include "da.h"

#include "isl_set.h"
#include "isl_map.h"

using pdg::PDG;
using namespace std;
using namespace da;

enum {
    TYPE_FLOW,
    TYPE_ANTI,
    TYPE_REUSE_PAIR,
    TYPE_OUTPUT,
};

const char *dep_opts[] = {
    "flow",
    "anti",
    "reuse_pair",
    "output",
    NULL
};

struct argp_option argp_options[] = {
    { "input",		    'i',		"file",	0 },
    { "output",		    'o',		"file",	0 },
    { "type",		    't',
	"flow,anti,reuse_pair,output", 0 },
    { "array-types",	    'A',		0,	0,
	"compute array types" },
    { "version",	    'V',	    	0,	0 },
    { 0 }
};

struct arguments {
    char *input;
    char *output;
    unsigned types;
    bool array_types;
};

error_t parse_opt(int key, char *arg, struct argp_state *state)
{
    struct arguments *arguments = (struct arguments *)(state->input);
    char *subopt;
    char *copy;

    switch (key) {
    case ARGP_KEY_INIT:
    	arguments->input = NULL;
    	arguments->output = NULL;
	arguments->types = (1 << TYPE_FLOW) | (1 << TYPE_ANTI);
	arguments->array_types = 0;
	break;
    case 'V':
	printf("%s\n", pdg_version());
	exit(0);
    case 'i':
	arguments->input = arg;
	break;
    case 'A':
	arguments->array_types = 1;
	break;
    case 'o':
	arguments->output = arg;
	break;
    case 't':
	arguments->types = 0;
	arg = copy = strdup(arg);
	while (*arg != '\0')
	    switch (getsubopt(&arg, (char* const*)dep_opts, &subopt)) {
	    case TYPE_FLOW:
		arguments->types |= (1 << TYPE_FLOW);
		break;
	    case TYPE_ANTI:
		arguments->types |= (1 << TYPE_ANTI);
		break;
	    case TYPE_REUSE_PAIR:
		arguments->types |= (1 << TYPE_REUSE_PAIR);
		break;
	    case TYPE_OUTPUT:
		arguments->types |= (1 << TYPE_OUTPUT);
		break;
	    default:
		argp_error(state, "unknown suboption '%s'\n", arg);
	    }
	free(copy);
	break;
    default:
	return ARGP_ERR_UNKNOWN;
    }
    return 0;
}

/* Return an input array that is otherwise a copy of a */
pdg::array *input_copy(PDG *pdg, pdg::array *a)
{
	pdg::array *copy = new pdg::array;
	copy->id = a->id;
	copy->name = a->name;
	copy->dims = a->dims;
	copy->map = a->map;
	copy->element_type = a->element_type;
	copy->type = pdg::array::input;
	pdg->arrays.push_back(copy);
	return copy;
}

/* Determine the type (input/output/temp) of each array in the program.
 * An array is considered
 *	- input, if some element is read without being written
 *	  If so, the corresponding access are changed to read from
 *	  a copy of the array marked as input.  The original array
 *	  is then marked temporary or output, as appropriate.
 *	- output, if some element is written without being subsequently read
 *	- temporary, otherwhise
 */
void determine_array_types(PDG *pdg)
{
	struct isl_ctx *ctx;
	std::map<pdg::node *, isl_set *> exposed;
	std::map<pdg::array *, pdg::array *> shadow_input;

	ctx = isl_ctx_alloc();
	assert(ctx);

	unsigned nparam = pdg->params.size();
	for (int j = 0; j < pdg->arrays.size(); ++j)
		pdg->arrays[j]->type = pdg::array::temp;
	for (int j = 0; j < pdg->dependences.size(); ++j) {
		pdg::dependence *dep = pdg->dependences[j];
		pdg::array *a = dep->array;
		if (dep->type != pdg::dependence::uninitialized)
			continue;
		if (shadow_input.find(a) == shadow_input.end())
			shadow_input[a] = input_copy(pdg, a);
		dep->array = shadow_input[a];
	}
	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		pdg::statement *s = node->statement;

		exposed[node] = node->source->get_isl_set(ctx);
	}
	for (int i = 0; i < pdg->dependences.size(); ++i) {
		pdg::dependence *dep = pdg->dependences[i];
		if (dep->type != pdg::dependence::flow)
			continue;
		pdg::node *node = dep->from;
		isl_map *rel = dep->relation->get_isl_map(ctx);
		exposed[node] = isl_set_subtract(ctx, exposed[node],
					isl_map_domain(ctx, rel));
	}
	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		pdg::statement *s = node->statement;

		int is_empty = isl_set_is_empty(ctx, exposed[node]);
		assert(is_empty >= 0);
		isl_set_free(ctx, exposed[node]);
		if (is_empty)
			continue;

		for (int j = 0; j < s->accesses.size(); ++j) {
			if (s->accesses[j]->type != pdg::access::write)
				continue;
			pdg::array *a = s->accesses[j]->array;
			a->type = pdg::array::output;
		}
	}
	isl_ctx_free(ctx);
}

int main(int argc, char * argv[])
{
    FILE *in = stdin, *out = stdout;
    PDG *pdg;
    bool output_allocated = false;
    struct arguments arguments;
    static struct argp argp = { argp_options, parse_opt, 0, 0, NULL };

    set_program_name(argv[0]);
    argp_parse(&argp, argc, argv, 0, 0, &arguments);

    if (arguments.input && strcmp(arguments.input, "-")) {
	in = fopen(arguments.input, "r");
	assert(in);
	if (!arguments.output) {
	    int len = strlen(arguments.input);
	    if (len > 5 && !strcmp(arguments.input+len-5, ".yaml"))
		len -= 5;
	    arguments.output = new char[len+9+1];
	    output_allocated = true;
	    strncpy(arguments.output, arguments.input, len);
	    strcpy(arguments.output+len, "_da.yaml");
	}
    }

    if (arguments.array_types)
	arguments.types |= 1 << TYPE_FLOW;

    pdg = yaml::Load<PDG>(in);
    assert(pdg);

    for (int i = 0; i < pdg->arrays.size(); ++i) {
	if (arguments.types & (1 << TYPE_FLOW))
	    find_deps(pdg, pdg->arrays[i], flow);
	if (arguments.types & (1 << TYPE_ANTI))
	    find_deps(pdg, pdg->arrays[i], anti);
	if (arguments.types & (1 << TYPE_REUSE_PAIR))
	    find_deps(pdg, pdg->arrays[i], reuse_pair);
	if (arguments.types & (1 << TYPE_OUTPUT))
	    find_deps(pdg, pdg->arrays[i], output);
    }

    if (arguments.array_types)
	determine_array_types(pdg);

    if (arguments.output && strcmp(arguments.output, "-")) {
	out = fopen(arguments.output, "w");
	assert(out);
    }
    if (output_allocated)
	delete [] arguments.output;

    pdg->add_history_line("da", argc, argv);

    pdg->Dump(out);
    pdg->free();
    delete pdg;

    return 0;
}
