#include <sstream>
#include "evalue_util.h"
#include <barvinok/util.h>

#define MAXRAYS    POL_NO_DUAL

#define ALLOC(type) (type*)malloc(sizeof(type))

void evalue_simplify_domains(evalue *e, Polyhedron *ctx)
{
    if (EVALUE_IS_ZERO(*e))
	return;
    assert(value_zero_p(e->d));
    assert(e->x.p->type == partition);
    for (int i = 0; i < e->x.p->size/2; ++i) {
	Polyhedron *D = EVALUE_DOMAIN(e->x.p->arr[2*i]);
	EVALUE_SET_DOMAIN(e->x.p->arr[2*i], DomainSimplify(D, ctx, MAXRAYS));
	Domain_Free(D);
    }
}

static void evalue_print(std::ostream& o, evalue *e, seq<pdg::parameter>&p);
static void evalue_print(std::ostream& o, evalue *e, seq<pdg::parameter>&p, int d)
{
    if (value_notzero_p(e->d)) {
	o << VALUE_TO_INT(e->x.n) * (d / VALUE_TO_INT(e->d));
	return;
    }
    assert(e->x.p->type == polynomial || e->x.p->type == flooring);
    int offset = e->x.p->type == flooring ? 1 : 0;
    for (int i = e->x.p->size-1; i >= offset; --i) {
	if (EVALUE_IS_ZERO(e->x.p->arr[i]))
	    continue;
	if (i != e->x.p->size-1 && 
		(value_zero_p(e->x.p->arr[i].d) ||
		value_pos_p(e->x.p->arr[i].x.n)))
	    o << "+";
	if (i == offset || !(value_one_p(e->x.p->arr[i].x.n) &&
			     d == VALUE_TO_INT(e->x.p->arr[i].d))) {
	    if (value_zero_p(e->x.p->arr[i].d))
		o << "(";
	    evalue_print(o, &e->x.p->arr[i], p, d);
	    if (value_zero_p(e->x.p->arr[i].d))
		o << ")";
	    if (i != offset)
		o << "*";
	}
	for (int j = 0; j < i-offset; ++j) {
	    if (j != 0)
		o << "*";
	    if (e->x.p->type == flooring) {
		o << "((int)(";
		evalue_print(o, &e->x.p->arr[0], p);
		o << "))";
	    } else
		o << p[e->x.p->pos-1]->name->s;
	}
    }
}

static void evalue_print(std::ostream& o, evalue *e, seq<pdg::parameter>&p)
{
    Value d;
    value_init(d);
    value_set_si(d, 1);
    evalue_denom(e, &d);
    if (value_notone_p(d))
	o << "(";
    evalue_print(o, e, p, VALUE_TO_INT(d));
    if (value_notone_p(d))
	o << ")/" << VALUE_TO_INT(d);
    value_clear(d);
}

static void evalue_printpoly(std::ostream& o, Polyhedron *D, seq<pdg::parameter>&p)
{
    for (int i = 0; i < D->NbConstraints; ++i) {
	int first = 1;
	if (i)
	    o << " && ";
	for (int j = 0; j < D->Dimension; ++j) {
	    if (value_zero_p(D->Constraint[i][1+j]))
		continue;
	    if (!first)
		o << " + ";
	    o << VALUE_TO_INT(D->Constraint[i][1+j]) << "*" << p[j]->name->s;
	    first = 0;
	}
	if (value_notzero_p(D->Constraint[i][1+D->Dimension])) {
	    if (!first)
		o << " + ";
	    o << VALUE_TO_INT(D->Constraint[i][1+D->Dimension]);
	}
	if (value_zero_p(D->Constraint[i][0]))
	    o << " == 0";
	else
	    o << " >= 0";
    }
}

static void evalue_printdomain(std::ostream& o, Polyhedron *D, 
			       seq<pdg::parameter>&p)
{
    if (!D->next)
	evalue_printpoly(o, D, p);
    else for (Polyhedron *P = D; P; P = P->next) {
	o << "(";
	evalue_printpoly(o, P, p);
	o << ")";
	if (P->next)
	    o << " || ";
    }
}

string evalue2string(evalue *e, seq<pdg::parameter>&p)
{
    std::ostringstream strm;
    if (EVALUE_IS_ZERO(*e))
	return "0";
    assert(value_zero_p(e->d));
    assert(e->x.p->type == partition);
    if (e->x.p->size == 2 && universeQ(EVALUE_DOMAIN(e->x.p->arr[0])))
	evalue_print(strm, &e->x.p->arr[1], p);
    else {
	for (int i = 0; i < e->x.p->size/2; ++i) {
	    strm << "(";
	    evalue_printdomain(strm, EVALUE_DOMAIN(e->x.p->arr[2*i]), p);
	    strm << ") ? (";
	    evalue_print(strm, &e->x.p->arr[2*i+1], p);
	    strm << ") :\n\t\t";
	}
	strm << "0";
    }
    return(strm.str());
}

/* add constraints to matrix starting from row specifying that
 * the n coordinates starting at pos1 are equal to those starting at pos2
 */
static void var_equal(Matrix *matrix, int pos1, int n, int pos2, int row)
{
    int i;
    for (i = 0; i < n; ++i) {
	Vector_Set(matrix->p[row+i], 0, matrix->NbColumns);
	value_set_si(matrix->p[row+i][0], 0);
	value_set_si(matrix->p[row+i][pos1+i], -1);
	value_set_si(matrix->p[row+i][pos2+i], 1);
	value_set_si(matrix->p[row+i][matrix->NbColumns-1], 0);
    }
}

/* add constraints to matrix starting from row specifying that
 * the coordinate at pos1 is smaller than the one at pos2
 */
static void var_smaller(Matrix *matrix, int pos1, int pos2, int row)
{
    Vector_Set(matrix->p[row], 0, matrix->NbColumns);
    value_set_si(matrix->p[row][0], 1);
    value_set_si(matrix->p[row][pos1], -1);
    value_set_si(matrix->p[row][pos2], 1);
    value_set_si(matrix->p[row][matrix->NbColumns-1], -1);
}

/* Computes the number of elements from P lexicographically between
 * pairs of elements from the relation R.
 *
 * All of P, R and C are treated as single polyhedra, i.e., the next
 * pointers are ignored.
 */
evalue *evalue_lexbetween_poly(Polyhedron *P, Polyhedron *R, unsigned dim,
			       Polyhedron *C, unsigned MaxRays, evalue *res)
{
    int nparam = C->Dimension;
    int extra_P = P->Dimension - dim - nparam;
    int extra_R = R->Dimension - dim - dim - nparam;
    int cols = 1 + dim + extra_P + dim + dim + extra_R + nparam + 1;
    int rows = P->NbConstraints + R->NbConstraints + C->NbConstraints + dim + dim;
    int row = 0;
    Matrix *M, *M2;
    Polyhedron *T;

    M = Matrix_Alloc(rows, cols);
    for (int i = 0; i < P->NbConstraints; ++i) {
	Vector_Copy(P->Constraint[i], M->p[i], 1+dim+extra_P);
	Vector_Copy(P->Constraint[i]+1+dim+extra_P,M->p[i]+cols-(nparam+1),nparam+1);
    }
    row += P->NbConstraints;
    for (int i = 0; i < R->NbConstraints; ++i) {
	value_assign(M->p[row+i][0], R->Constraint[i][0]);
	Vector_Copy(R->Constraint[i]+1, M->p[row+i]+1+dim+extra_P, R->Dimension+1);
    }
    row += R->NbConstraints;
    for (int i = 0; i < C->NbConstraints; ++i) {
	value_assign(M->p[row+i][0], C->Constraint[i][0]);
	Vector_Copy(C->Constraint[i]+1, M->p[row+i]+cols-(nparam+1), nparam+1);
    }
    row += C->NbConstraints;

    for (int dW = 0; dW < dim; ++dW) {
	var_equal(M, 1+dim+extra_P, dW, 1, row);
	var_smaller(M, 1+dim+extra_P+dW, 1+dW, row+dW);
	for (int dR = 0; dR < dim; ++dR) {
	    evalue *t;

	    var_equal(M, 1, dR, 1+dim+extra_P+dim, row+dW+1);
	    var_smaller(M, 1+dR, 1+dim+extra_P+dim+dR, row+dW+1+dR);
	    M->NbRows = row+dW+1+dR+1;
	    M2 = Matrix_Copy(M);
	    T = Constraints2Polyhedron(M2, MAXRAYS);

	    t = barvinok_enumerate_e(T, extra_P+dim, dim+extra_R+nparam, MaxRays);

	    if (!res)
		res = t;
	    else if (t) {
		eadd(t, res);
		free_evalue_refs(t);
		free(t);
	    }

	    Polyhedron_Free(T);
	    Matrix_Free(M2);
	}
    }
    Matrix_Free(M);
    return res;
}

/* Like barvinok_lexsmaller_ev, except that D may be a union of polyhedra */
evalue *lexsmaller_domain(Polyhedron *D, Polyhedron *V, int dim, Polyhedron *ctx)
{
    evalue *count = barvinok_lexsmaller_ev(D, V, dim, ctx, MAXRAYS);
    for (D = D->next; D; D = D->next) {
	evalue *t = barvinok_lexsmaller_ev(D, V, dim, ctx, MAXRAYS);
	eadd(t, count);
	free_evalue_refs(t);
	free(t);
    }
    return count;
}

static void evalue_ap(evalue *e, int *active);
static void enode_ap(enode *p, int *active)
{
    switch (p->type) {
    case polynomial:
	active[p->pos-1] = 1;
    case flooring:
    case fractional:
    case relation:
	for (int i = 0; i < p->size; ++i)
	    evalue_ap(&p->arr[i], active);
	break;
    default:
	assert(0);
    }
}

static void evalue_ap(evalue *e, int *active)
{
    if (value_notzero_p(e->d))
	return;
    enode_ap(e->x.p, active);
}

int evalue_active_params(evalue *e, int *active, int dim)
{
    for (int i = 0; i < dim; ++i)
	active[i] = 0;
    evalue_ap(e, active);
    for (int i = 0; i < dim; ++i)
	if (!active[i])
	    return 1;
    return 0;
}

/* test whether e is symbolic constant, i.e. independent of first dim dimensions */
bool evalue_is_constant(evalue *e, int dim)
{
    if (EVALUE_IS_ZERO(*e))
	return true;
    assert(value_zero_p(e->d));
    assert(e->x.p->type == partition);
    int totdim = EVALUE_DOMAIN(e->x.p->arr[0])->Dimension;
    int active[totdim];
    for (int i = 0; i < e->x.p->size/2; ++i) {
	evalue_active_params(&e->x.p->arr[2*i+1], active, totdim);
	for (int j = 0; j < dim; ++j)
	    if (active[j])
		return false;
    }
    return true;
}

void evalue_remove_iterators(evalue *EP, Polyhedron *P, int dim, Polyhedron *ctx)
{
    int nparam = ctx->Dimension;
    Value mone;
    value_init(mone);
    value_set_si(mone, -1);
    Matrix *M = Matrix_Alloc(dim+nparam+1, nparam+1);
    for (int j = 0; j < dim; ++j)
	Vector_Scale(P->Constraint[j]+dim+1, M->p[j], mone, nparam+1);
    for (int i = 0; i < nparam+1; ++i)
	value_set_si(M->p[dim+i][i], 1);
    Polyhedron *U = Universe_Polyhedron(nparam);
    /* We're actually _removing_ parameters here, but it should work */
    addeliminatedparams_enum(EP, M, U, MAXRAYS, nparam);
    evalue_simplify_domains(EP, ctx);
    Matrix_Free(M);
    Polyhedron_Free(U);
    value_clear(mone);
}

/* returns the evalue D |-> 1 */
evalue *evalue_domain(Polyhedron *D)
{
    evalue *res = ALLOC(evalue);

    value_init(res->d);
    value_set_si(res->d, 0);
    res->x.p = new_enode(::partition, 2, D->Dimension);
    EVALUE_SET_DOMAIN(res->x.p->arr[0], D);
    evalue_set_si(&res->x.p->arr[1], 1, 1);
    return res;
}
