#include <vector>
#include <bernstein/bernstein.h>
#include <bernstein/piecewise_lst.h>
#include <barvinok/bernstein.h>
#include <barvinok/util.h>
#include <ginac/ginac.h>
#include "argp.h"
#include "yaml.h"
#include "pdg.h"
#include "da.h"
#include "evalue_util.h"
#include "translation.h"
extern "C" {
#include "piputil.h"
}

using std::vector;
using pdg::PDG;
using namespace da;
using namespace GiNaC;
using namespace barvinok;
using namespace bernstein;
using std::cout;
using std::cerr;
using std::endl;

#define OPT_WITH_INPUT	    (BV_OPT_LAST+1)

struct argp_option argp_options[] = {
    { "with-input",	OPT_WITH_INPUT,		0,	0,
	"include size of input arrays in total" },
    { 0 }
};

struct options {
    barvinok_options	*barvinok;
    int			 with_input;
};

static error_t parse_opt(int key, char *arg, struct argp_state *state)
{
    struct options *options = (struct options *)(state->input);

    switch (key) {
    case ARGP_KEY_INIT:
	state->child_inputs[0] = options->barvinok;
	options->with_input = 0;
	break;
    case OPT_WITH_INPUT:
	options->with_input = 1;
	break;
    default:
	return ARGP_ERR_UNKNOWN;
    }
    return 0;
}

static exvector construct_params(pdg::PDG *pdg)
{
    int nparam = pdg->params.size();
    const char *param_names[nparam];
    for (int i = 0; i < nparam; ++i)
	param_names[i] = pdg->params[i]->name->s.c_str();
    return constructParameterVector(param_names, nparam);
}

/* Compute the number of array elements that are read but not
 * (previously) written.  These are assumed to be the input.
 */
static evalue *compute_input_size(PDG *pdg, struct options *options)
{
    if (!options->with_input)
	return evalue_zero();

    unsigned nparam = pdg->params.size();
    evalue *inputsize = evalue_zero();
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type != pdg::dependence::uninitialized)
	    continue;
	pdg::UnionMap *rel = dep->relation;
	Polyhedron *D = *rel;
	for (Polyhedron *P = D; P; P = P->next) {
	    evalue *EP = barvinok_enumerate_e_with_options(P,
			    P->Dimension - rel->output - nparam, nparam,
			    options->barvinok);
	    eadd(EP, inputsize);
	    evalue_free(EP);
	}
	Domain_Free(D);
    }
    return inputsize;
}

/* Reorder the variables in "D" such that the first "nvar" variables
 * are place right before the existing parameters.
 * The order of the remaining variables is kept.
 */
static void make_first_vars_params(Polyhedron *D, int nvar, int nparam, int n)
{
    assert(n <= nvar);
    for (Polyhedron *P = D; P; P = P->next) {
	int exist = P->Dimension - nparam - nvar;
	Vector *V = Vector_Alloc(nvar+exist);
	for (int i = 0; i < P->NbConstraints; ++i) {
	    Vector_Copy(P->Constraint[i]+1, V->p+nvar+exist-n, n);
	    Vector_Copy(P->Constraint[i]+1+n, V->p, nvar-n);
	    Vector_Copy(P->Constraint[i]+1+nvar, V->p+nvar-n, exist);
	    Vector_Copy(V->p, P->Constraint[i]+1, nvar+exist);
	}
	for (int i = 0; i < P->NbRays; ++i) {
	    Vector_Copy(P->Ray[i]+1, V->p+nvar+exist-n, n);
	    Vector_Copy(P->Ray[i]+1+n, V->p, nvar-n);
	    Vector_Copy(P->Ray[i]+1+nvar, V->p+nvar-n, exist);
	    Vector_Copy(V->p, P->Ray[i]+1, nvar+exist);
	}
	if (P->NbEq) {
	    Matrix M;
	    Polyhedron_Matrix_View(P, &M, P->NbConstraints);
	    Gauss(&M, P->NbEq, P->Dimension+1);
	}
	Vector_Free(V);
    }
}

/*
 * Create a new domain with an extra variable right before
 * the parameters in D.
 */
static Polyhedron *insert_parameter(Polyhedron *D, int nvar, int nparam,
					int level, barvinok_options *options)
{
    Polyhedron *result = NULL;
    Polyhedron **next = &result;

    for (Polyhedron *P = D; P; P = P->next) {
	int exist = P->Dimension - nparam - nvar;
	Matrix *M = Matrix_Alloc(P->NbConstraints+1, 1+P->Dimension+1+1);
	for (int i = 0; i < P->NbConstraints; ++i) {
	    Vector_Copy(P->Constraint[i], M->p[i+1], 1+nvar+exist);
	    Vector_Copy(P->Constraint[i]+1+nvar+exist,
			M->p[i+1]+1+nvar+exist+1, nparam+1);
	}
	value_set_si(M->p[0][0], 1);
	value_set_si(M->p[0][1+level], -1);
	value_set_si(M->p[0][1+nvar+exist], 1);
	value_set_si(M->p[0][M->NbColumns-1], -1);
	Polyhedron *Q = Constraints2Polyhedron(M, options->MaxRays);

	Matrix_Free(M);
	*next = Q;
	next = &Q->next;
    }
    return result;
}

/* Compute the number of elements in domain for a fixed value
 * of the "fixed_loops" outer loop iterators and return this value
 * as a function of these loop iterators and the parameters.
 */
static evalue *compute_accesses_at(Polyhedron *domain, int nvar, int nparam,
				int fixed_loops, barvinok_options *options)
{
    Polyhedron *D = Domain_Copy(domain);
    make_first_vars_params(D, nvar, nparam, fixed_loops);

    evalue *sum = evalue_zero();
    for (Polyhedron *P = D; P; P = P->next) {
	evalue *e = barvinok_enumerate_e_with_options(P,
			P->Dimension - nvar - nparam, fixed_loops + nparam,
			options);
	eadd(e, sum);
	evalue_free(e);
    }
    Domain_Free(D);

    return sum;
}

/* Compute the number of elements in domain for a fixed value
 * of the "level" outer loop iterators and for all values of the
 * loop iterator at level "level" smaller than a fixed value
 * and return this value as a function of these (level+1) loop iterators
 * and the parameters.
 */
static evalue *compute_accesses_before(Polyhedron *domain, int nvar, int nparam,
				int level, barvinok_options *options)
{
    Polyhedron *D = insert_parameter(domain, nvar, nparam, level, options);
    ++nparam;
    make_first_vars_params(D, nvar, nparam, level);

    evalue *sum = evalue_zero();
    for (Polyhedron *P = D; P; P = P->next) {
	evalue *e = barvinok_enumerate_e_with_options(P,
			P->Dimension - nvar - nparam, level + nparam,
			options);
	eadd(e, sum);
	evalue_free(e);
    }
    Domain_Free(D);
    return sum;
}

static int cmp_prefix_stat_dims(pdg::node *node, int *stat_dims, int d)
{
    for (int j = 0; j < d; ++j)
	if (node->prefix[2*j] != stat_dims[j])
	    return node->prefix[2*j] - stat_dims[j];

    return 0;
}

/*
 * Returns true if dependence is independent of first n_loops loops.
 */
static int is_loop_independent(pdg::UnionMap *relation, int n_loops)
{
    assert(relation->input >= n_loops);
    assert(relation->output >= n_loops);

    for (int i = 0; i < n_loops; ++i) {
	for (int j = 0; j < relation->constraints.size(); ++j) {
	    if (relation->constraints[j]->el.size() <= i)
		return 0;
	    if (relation->constraints[j]->el[i][0] != 0)
		return 0;
	    int n = relation->constraints[j]->el[i].size()-1;
	    for (int k = 0; k < n; ++k) {
		int v = relation->constraints[j]->el[i][1+k];
		if (k == i) {
		    if (v != 1)
			return 0;
		} else if (k == relation->input+i) {
		    if (v != -1)
			return 0;
		} else
		    if (v != 0)
			return 0;
	    }
	}
    }

    return 1;
}

struct contribution {
    unsigned	is_read : 1;
    unsigned	is_written : 1;
};

/*
 * Determines the net contribution of "dep" within statements
 * with statement level dimensions prefixed by stat_dims
 * and within a fixed iteration of the first n_loop loops.
 */
static struct contribution net_contribution(pdg::dependence *dep,
					    int *stat_dims, int d, int n_loop,
					    struct options *options)
{
    struct contribution c;
    c.is_read = 0;
    c.is_written = 0;

    if (dep->type == pdg::dependence::uninitialized) {
	if (options->with_input &&
		cmp_prefix_stat_dims(dep->to, stat_dims, d) == 0)
	    c.is_read = 1;
    } else {
	int is_written = cmp_prefix_stat_dims(dep->from, stat_dims, d) == 0;
	int is_read = cmp_prefix_stat_dims(dep->to, stat_dims, d) == 0;
	int is_local = is_written && is_read &&
				is_loop_independent(dep->relation, n_loop);
	if (!is_local) {
	    c.is_read = is_read;
	    c.is_written = is_written;
	}
    }

    return c;
}

static evalue *evalue_diff(evalue *a, evalue *b)
{
    evalue_negate(b);
    if (!a)
	return b;
    eadd(b, a);
    reduce_evalue(a);
    evalue_free(b);
    return a;
}

struct access {
    Polyhedron	*read;
    Polyhedron	*write;
};

/*
 * Compute increment in statement/loop identified by stat_dims
 */
static evalue *compute_live_increment(PDG *pdg, int *stat_dims, int d,
					struct access *accesses,
					struct options *options)
{
    evalue *sum = evalue_zero();
    unsigned nparam = pdg->params.size();
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	evalue *inc = NULL;
	pdg::dependence *dep = pdg->dependences[i];
	struct contribution c;

	c = net_contribution(dep, stat_dims, d, d-1, options);

	if (c.is_written) {
	    inc = compute_accesses_at(accesses[i].write,
					dep->relation->input, nparam, d-1,
					options->barvinok);
	}
	if (c.is_read) {
	    evalue *dec;
	    dec = compute_accesses_at(accesses[i].read,
					dep->relation->output, nparam, d-1,
					options->barvinok);
	    inc = evalue_diff(inc, dec);
	}
	if (inc) {
	    eadd(inc, sum);
	    evalue_free(inc);
	}
    }
    return sum;
}

/*
 * Compute increment in previous iterations of loop at given level.
 */
static evalue *compute_live_increment_in_loop(PDG *pdg, int *stat_dims,
					int level,
					struct access *accesses,
					struct options *options)
{
    evalue *sum = evalue_zero();
    unsigned nparam = pdg->params.size();
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	evalue *inc = NULL;
	pdg::dependence *dep = pdg->dependences[i];
	struct contribution c;

	c = net_contribution(dep, stat_dims, level+1, level+1, options);

	if (c.is_written) {
	    inc = compute_accesses_before(accesses[i].write,
				    dep->relation->input, nparam,
				    level, options->barvinok);
	}
	if (c.is_read) {
	    evalue *dec;
	    dec = compute_accesses_before(accesses[i].read,
				    dep->relation->output, nparam,
				    level, options->barvinok);
	    inc = evalue_diff(inc, dec);
	}
	if (inc) {
	    eadd(inc, sum);
	    evalue_free(inc);
	}
    }
    return sum;
}

struct tree;

struct tree {
    int d;
    int *stat_dims;
    vector<struct tree*> children;
    Polyhedron	 	*domain;
};

/*
 * Add a new variable after the current variables (before the parameters).
 */
static void evalue_append_variable(evalue *e, unsigned nvar, unsigned nparam,
					unsigned MaxRays)
{
    Matrix *M = Matrix_Alloc(nvar+nparam+1, nvar+1+nparam+1);
    for (int i = 0; i < nvar; ++i)
	value_set_si(M->p[i][i], 1);
    for (int i = 0; i < nparam+1; ++i)
	value_set_si(M->p[nvar+i][nvar+1+i], 1);
    addeliminatedparams_enum(e, M, NULL, MaxRays, nvar+1+nparam);
    Matrix_Free(M);
}

/*
 * Compute bound on number of live array elements in the
 * given (sub)tree.
 *
 * live_before: number of live elements before the loop/statement
 *		in terms of the outer loop iterators
 */
static piecewise_lst *compute_bound(PDG *pdg,
					piecewise_lst *pl,
					const exvector& params,
					Polyhedron *ctx,
					const evalue *live_before,
					struct tree *tree,
					struct access *accesses,
					struct options *options)
{
    /* If we are in a leaf, we actually compute the bound. */
    if (tree->children.size() == 0) {
	evalue *mask = evalue_domain(tree->domain);
	tree->domain = NULL;
	emul(live_before, mask);
	pl = evalue_bernstein_coefficients(pl, mask, ctx, params,
						options->barvinok);
	evalue_free(mask);
	return pl;
    }

    evalue *live = evalue_dup(live_before);

    /* If we are not in the outer dummy loop, add contribution of
     * previous iterations.
     */
    if (tree->d > 0) {
	unsigned nparam = pdg->params.size();
	evalue_append_variable(live, tree->d-1, nparam,
				options->barvinok->MaxRays);

	evalue *inc;
	inc = compute_live_increment_in_loop(pdg, tree->stat_dims, tree->d-1,
					    accesses, options);
	eadd(inc, live);
	evalue_free(inc);
    }

    pl = compute_bound(pdg, pl, params, ctx,
			    live, tree->children[0], accesses, options);
    for (int i = 1; i < tree->children.size(); ++i) {
	struct tree *child = tree->children[i-1];
	evalue *inc;
	inc = compute_live_increment(pdg, child->stat_dims, child->d,
				    accesses, options);
	eadd(inc, live);
	evalue_free(inc);
	pl = compute_bound(pdg, pl, params, ctx,
				live, tree->children[i], accesses, options);
    }
    evalue_free(live);
    return pl;
}

/*
 * Construct a tree for all statements whose statement dimensions
 * start with the d first elements of stat_dims.
 * The first one of those statemens is statement n.
 */
static struct tree *compute_tree(PDG *pdg, int *stat_dims, int d, int n)
{
    struct tree *tree = new struct tree;
    tree->domain = NULL;
    tree->d = d;
    tree->stat_dims = new int[d];
    for (int i = 0; i < d; ++i)
	tree->stat_dims[i] = stat_dims[i];

    if (pdg->nodes[n]->prefix.size() == 2*d-1) {
	tree->domain = *pdg->nodes[n]->source;
	return tree;
    }

    for (int i = n; i < pdg->nodes.size(); ++i) {
	if (cmp_prefix_stat_dims(pdg->nodes[i], stat_dims, d) > 0)
	    break;

	stat_dims[d] = pdg->nodes[i]->prefix[2*d];

	struct tree *child = compute_tree(pdg, stat_dims, d+1, i);
	tree->children.push_back(child);

	/* Skip over nodes with the same prefix at current level */
	for (; i+1 < pdg->nodes.size(); ++i)
	    if (cmp_prefix_stat_dims(pdg->nodes[i+1], stat_dims, d+1) != 0)
		break;
    }
    return tree;
}

static struct tree *compute_tree(PDG *pdg)
{
    int n_stat_dims = normalize_prefix(pdg);
    int stat_dims[n_stat_dims];

    return compute_tree(pdg, stat_dims, 0, 0);
}

static void free_tree(struct tree *tree)
{
    for (int i = 0; i < tree->children.size(); ++i)
	free_tree(tree->children[i]);
    delete [] tree->stat_dims;
    if (tree->domain)
	Domain_Free(tree->domain);
    delete tree;
}

static void compute_bound(PDG *pdg, const exvector& params, Polyhedron *ctx,
				struct options *options)
{
    options->barvinok->bernstein_optimize = BV_BERNSTEIN_MAX;

    struct tree *tree = compute_tree(pdg);

    struct access accesses[pdg->dependences.size()];
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];

	accesses[i].write = *dep->relation;

	Polyhedron *reads = NULL;
	for (Polyhedron *P = accesses[i].write; P; P = P->next) {
	    Polyhedron *read = Polyhedron_projectout_dep(P, dep->relation->input);
	    reads = Domain_concat(read, reads);
	}

	accesses[i].read = reads;
    }

    evalue *inputsize = compute_input_size(pdg, options);

    piecewise_lst *pl;
    pl = compute_bound(pdg, NULL, params, ctx,
				inputsize, tree, accesses, options);
    evalue_free(inputsize);
    free_tree(tree);
    pl->maximize();
    pl->simplify_domains(ctx, options->barvinok->MaxRays);
    cout << *pl << endl;
    delete pl;

    for (int i = 0; i < pdg->dependences.size(); ++i) {
	Domain_Free(accesses[i].write);
	Domain_Free(accesses[i].read);
    }
}

int main(int argc, char * argv[])
{
    PDG *pdg;
    static struct argp_child argp_children[] = {
	{ &barvinok_argp,    	0,	"barvinok options",	1 },
	{ 0 }
    };
    static struct argp argp = { argp_options, parse_opt, 0, 0, argp_children };
    struct barvinok_options *bv_options = barvinok_options_new_with_defaults();
    struct options options;

    options.barvinok = bv_options;
    argp_parse(&argp, argc, argv, 0, 0, &options);

    pdg = yaml::Load<PDG>(stdin);
    int nparam = pdg->params.size();

    Polyhedron *ctx = pdg->context ? (Polyhedron *) *pdg->context : 
				     Universe_Polyhedron(nparam);

    for (int i = 0; i < pdg->arrays.size(); ++i)
	find_deps(pdg, pdg->arrays[i], data_reuse);

    exvector params = construct_params(pdg);

    compute_bound(pdg, params, ctx, &options);
    Polyhedron_Free(ctx);

    pdg->free();
    delete pdg;
    barvinok_options_free(bv_options);

    return 0;
}
