#ifndef PDG_H
#define PDG_H

#include "yaml.h"
#include <gmp.h>
#define matrix	poly_matrix
extern "C" {
#include <polylib/polylibgmp.h>
}
#undef matrix
#undef value_compare

struct isl_ctx;
struct isl_set;
struct isl_map;

using namespace yaml;

namespace pdg
{

class Matrix : public structure {
    static serialize *create() { return new Matrix(); }

public:
    std::vector< std::vector<int> > el;
    static void register_type();
    void init(SyckParser *p, SyckNode *n);
    void dump();
    void dump(emitter& e);
    Matrix() {};
    Matrix(Polyhedron *P);
    Matrix(::Matrix *M);
    operator ::Matrix *() const;
};

Polyhedron *constraints2polyhedron(seq<pdg::Matrix> const & c);

class parameter : public structure {
    static serialize *create() { return new parameter(); }
public:
    int	    id;
    str	    *name;
    integer *value;

    static void register_type();
    parameter () : id(-1), name(NULL), value(NULL) {}
};

class UnionSet : public structure {
    static serialize *create() { return new UnionSet(); }

public:
    int	    	    dim;
    seq<Matrix>	    constraints;
    seq<parameter>  params;

    static void register_type();
    void dump(emitter& e);
    UnionSet () : dim(-1) {}
    UnionSet(int d, Polyhedron *P, seq<parameter>& p);
    operator ::Polyhedron *() const;
    isl_set *get_isl_set(isl_ctx *ctx) const;
    void fill_fields(anchor_map *);
};

class UnionMap : public structure {
    static serialize *create() { return new UnionMap(); }

public:
    int	    input;
    int	    output;
    seq<Matrix>	constraints;
    seq<parameter>  params;

    static void register_type();
    void dump(emitter& e);
    UnionMap (int in = -1, int out = -1) 
	: input(in), output(out) {}
    UnionMap(int in, int out, Polyhedron *P, seq<parameter>& p);
    operator ::Polyhedron *() const;
    isl_map *get_isl_map(isl_ctx *ctx) const;
    void fill_fields(anchor_map *);
};

class dependence_type;

class array : public structure {
    static serialize *create() { return new array(); }
public:
    int	    id;
    str	    *name;
    seq<int> dims;
    UnionMap *    map;
    str	    *element_type;
    enum type {
	unset = -1,
	input,
	output,
	temp,
	last
    } type;
    seq<dependence_type>	analysis_performed;

    static void register_type();
    array () : id(-1), name(NULL), map(NULL), element_type(NULL), type(unset) {}
};

class ComposedMap : public structure {
    static serialize *create() { return new ComposedMap(); }
public:
    seq<UnionMap>   pos;
    seq<UnionMap>   neg;
    seq<UnionSet>   mask;

    static void register_type();
    ComposedMap () {}
};

class access;

class reuse_pair : public structure {
    static serialize *create() { return new reuse_pair(); }
public:
    access *		from;
    UnionMap *		map;
    seq<ComposedMap>	ads;
    str *		ads_file;
    int			count;

    static void register_type();
    reuse_pair () : from(NULL), map(NULL), ads_file(NULL), count(-1) {}
};

class access : public structure {
    static serialize *create() { return new access(); }
public:
    pdg::array *	array;
    UnionMap *	map;
    enum type {
	unset = -1,
	write,
	read,
	last,
    } type;
    seq<reuse_pair>	backward;
    int		nr;

    static void register_type();
    access () : array(NULL), type(unset), map(NULL), nr(-1) {}
};

class function_call;

class call_or_access : public structure {
    static serialize *create() { return new call_or_access(); }
public:
    enum type {
	unset = -1,
	t_access,
	t_call,
	last,
    } type;
    pdg::access		*access;
    pdg::function_call	*call;

    static void register_type();
    call_or_access() : type(unset), access(NULL), call(NULL) {}
};

class function_call : public structure {
    static serialize *create() { return new function_call(); }
public:
    str			*name;
    seq<call_or_access>	 arguments;

    static void register_type();
    function_call() : name(NULL) {}
};

class statement : public structure {
    static serialize *create() { return new statement(); }
public:
    seq<access> accesses;
    int	    	operation;
    int		line;
    function_call	*top_function;

    static void register_type();
    statement () : operation(-1), top_function(NULL) {}
};

class dependence;

class node : public structure {
    static serialize *create() { return new node(); }
public:
    int	    	nr;
    UnionSet   	*source;
    UnionMap  	*scattering;
    pdg::statement	*statement;
    seq<int>		 prefix;
    seq<dependence>	 sources;

    static void register_type();
    node () : nr(-1), source(NULL), statement(NULL), scattering(NULL) {}
};

pdg::UnionMap *scatter_dependence(pdg::dependence *dep);

class enumerator : public structure {
    static serialize *create() { return new enumerator(); }
public:
    int		value;
    string     *s;

    static void register_type();
    void init(SyckParser *p, SyckNode *n);
    void dump(emitter& e);
    void operator= (int i);
    enumerator() : value(-1), s(NULL) {}
    enumerator(int v) : value(v), s(NULL) {}
    enumerator(string *s) : value(-1), s(s) {}
    ~enumerator();
};

class dependence : public structure {
    static serialize *create() { return new dependence(); }
public:
    UnionMap *relation;
    node    *from;
    node    *to;
    access    *from_access;
    access    *to_access;
    enum type {
	unset = -1,
	flow,
	anti,
	pn,
	pn_union,
	pn_part,
	pn_wire,
	pn_hold,
	pn_shift,
	reuse,
	reuse_pair,
	output,
	uninitialized,
	last,
    } type;
    pdg::array *array;
    int		reordering;
    int		multiplicity;
    pdg::enumerator *	size;
    pdg::dependence *	container;
    integer *	value_size;

    static void register_type();
    dependence () : relation(NULL), from(NULL), to(NULL), 
		    from_access(NULL), to_access(NULL), type(unset),
		    array(NULL), reordering(-1), multiplicity(-1),
		    size(NULL), container(NULL),
		    value_size(NULL) {}
};

class dependence_type : public structure {
public:
    static serialize *create() { return new dependence_type(); }
    dependence_type() : type(pdg::dependence::unset) {}
    dependence_type(enum pdg::dependence::type type) : type(type) {}
    enum pdg::dependence::type	type;
};

class history_entry : public structure {
    static serialize *create() { return new history_entry(); }
public:
    str	    *tool;
    str	    *version;
    seq<str>	arguments;

    static void register_type();
    history_entry(): tool(NULL), version(NULL) {}
};

class PDG : public structure {
    static serialize *create() { return new PDG(); }
public:
    int	    dimension;
    int	    cacheline;
    int	    root;
    str	    *placement;
    seq<node>  nodes;
    seq<array> arrays;
    seq<dependence> dependences;
    seq<parameter>  params;
    UnionSet   	*context;
    seq<history_entry>	history;
    str		*name;
    seq<int>	 statement_dimensions;

    static void register_type();
    void dump(emitter& e);
    PDG () : dimension(0), root(-1), cacheline(-1), placement(NULL), context(NULL),
	     name(NULL) {}

    void add_history_line(const char *tool, int argc, char * argv[]);
};

class HMatrix : public structure {
    static serialize *create() { return new Matrix(); }

public:
    static void register_type();
    void init(SyckParser *p, SyckNode *n) {};
};

}

#endif
