#include <sstream>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <dlfcn.h>
#include <gmp.h>
#include <barvinok/barvinok.h>
#include <barvinok/util.h>
#undef value_compare
#include "size.h"
#include "size_bernstein.h"
#include "evalue_util.h"
extern "C" {
#include <polylib/polylibgmp.h>
#include "combine.h"
#include "piputil.h"
}
#include <cloog/cloog.h>
#ifdef HAVE_BERNSTEIN
#include <cln/cln.h>
#include <ginac/ginac.h>
#endif
#include "clast_scanner.h"
#include "cloog_util.h"

namespace size {

using std::cerr;
using std::endl;
using std::ostream;

#ifndef HAVE_BERNSTEIN
size::enumerator *selfloop_size_bernstein(pdg::PDG *pdg, Polyhedron *dep,
					  Polyhedron *reads,
					  int dim, Polyhedron *max,
					  Polyhedron *ctx,
					  size_options *options)
{
    return NULL;
}
#endif

static void evalue_max_relation(evalue *e);

static int lexsmaller_point(Polyhedron *D, Polyhedron *V, int dim)
{
    int c;
    Polyhedron *U = Universe_Polyhedron(0);
    evalue *count = lexsmaller_domain(D, V, dim, U);
    Polyhedron_Free(U);
    if (EVALUE_IS_ZERO(*count))
	c = 0;
    else {
	assert(value_zero_p(count->d));
	assert(count->x.p->type == partition);
	assert(count->x.p->size == 2);
	assert(value_one_p(count->x.p->arr[1].d));
	c = VALUE_TO_INT(count->x.p->arr[1].x.n);
    }
    free_evalue_refs(count);
    free(count);
    return c;
}

static int live(Polyhedron *W, Polyhedron *R, Polyhedron *P, 
		int n_write, int n_read, unsigned MaxRays)
{
    Matrix *M = Matrix_Alloc(n_write, n_write+2);
    int l = 0;

    for ( ; P; P = P->next) {
	POL_ENSURE_VERTICES(P);
	for (int i = 0; i < P->NbRays; ++i)
	    if (!value_one_p(P->Ray[i][1+P->Dimension])) {
		l = -1;
		goto out;
	    }
	for (int i = 0; i < P->NbRays; ++i) {
	    for (int j = 0; j < n_write; ++j) {
		value_set_si(M->p[j][1+j], -1);
		value_assign(M->p[j][1+n_write], P->Ray[i][1+j]);
	    }
	    Polyhedron *V = Constraints2Polyhedron(M, MaxRays);
	    int p = lexsmaller_point(W, V, n_write);
	    int c = lexsmaller_point(R, V, n_write);
	    Polyhedron_Free(V);
	    assert(p >= c);

	    if (p - c > l)
		l = p - c;
	}
    }
out:
    Matrix_Free(M);
    return l;
}

static int evalue_max_on(evalue *e, Polyhedron *S, int pos, Value *z, int max)
{
    Value LB, UB;
    value_init(LB); value_init(UB);
    int ok = !(lower_upper_bounds(1+pos, S, z, &LB, &UB));
    assert(ok);

    for ( ; value_le(LB, UB); value_increment(LB, LB)) {
	value_assign(z[pos+1], LB);
	if (S->next)
	    max = evalue_max_on(e, S->next, pos+1, z, max);
	else {
	    int size = (int)(compute_evalue(e, z+1)+.25);
	    if (size > max)
		max = size;
	}
    }
    value_set_si(z[pos+1], 0);
    value_clear(LB); value_clear(UB);
    return max;
}

static int evalue_max_on(evalue *e, Polyhedron *P, unsigned MaxRays)
{
    Polyhedron *U = Universe_Polyhedron(0);
    Polyhedron *S = Polyhedron_Scan(P, U, MaxRays);
    Vector *z = Vector_Alloc(1+P->Dimension+1);
    value_set_si(z->p[1+P->Dimension], 1);
    int size = evalue_max_on(e, S, 0, z->p, -1);
    Vector_Free(z);
    Domain_Free(S);
    Polyhedron_Free(U);
    return size;
}

static Polyhedron *active_domain(Polyhedron *D, int *active, unsigned MaxRays)
{
    int n = 0;
    for (int i = 0; i < D->Dimension; ++i)
	if (!active[i])
	    ++n;
    Matrix *lines = Matrix_Alloc(n, 2+D->Dimension);
    for (int i = 0, j = 0; i < D->Dimension; ++i) {
	if (active[i])
	    continue;
	value_set_si(lines->p[j][1+i], 1);
	++j;
    }
    Polyhedron *D1 = DomainAddRays(D, lines, MaxRays);
    Polyhedron *D2 = DomainAddConstraints(D1, lines, MaxRays);
    Matrix_Free(lines);
    Domain_Free(D1);
    return D2;
}

static int evalue_max(evalue *e, unsigned MaxRays)
{
    if (EVALUE_IS_ZERO(*e))
	return 0;
    assert(value_zero_p(e->d));
    assert(e->x.p->type == partition);
    int maxsize = -1;
    int dim = EVALUE_DOMAIN(e->x.p->arr[0])->Dimension;
    int active[dim];
    for (int i = 0; i < e->x.p->size/2; ++i) {
	if (value_notzero_p(e->x.p->arr[2*i+1].d)) {
	    assert(value_one_p(e->x.p->arr[2*i+1].d));
	    int size = VALUE_TO_INT(e->x.p->arr[2*i+1].x.n);
	    if (size > maxsize)
		maxsize = size;
	    continue;
	}

	Polyhedron *D = EVALUE_DOMAIN(e->x.p->arr[2*i]);
	Polyhedron *E = D;
	if (evalue_active_params(&e->x.p->arr[2*i+1], active, dim))
	    E = active_domain(D, active, MaxRays);
	for (Polyhedron *P = E; P; P = P->next) {
	    Polyhedron *next = P->next;
	    P->next = NULL;
	    int size = evalue_max_on(&e->x.p->arr[2*i+1], P, MaxRays);
	    if (size > maxsize)
		maxsize = size;
	    P->next = next;
	}
	if (D != E)
	    Domain_Free(E);
    }
    return maxsize;
}

static bool is_rectangular_dependence(Polyhedron *dep, 
				      int n_write, int n_read, int nparam)
{
    if (dep->next)
	return false;
    for (int i = 0; i < n_write; ++i) {
	if (value_notzero_p(dep->Constraint[i][0]))
	    return false;
	if (value_notone_p(dep->Constraint[i][1+i]))
	    return false;
	if (value_zero_p(dep->Constraint[i][1+n_write+i])) {
	    Value mone;
	    value_init(mone);
	    value_set_si(mone, -1);
	    Vector_Combine(dep->Constraint[i]+1+n_write,
			   dep->Constraint[i+n_write]+1+n_write,
			   dep->Constraint[i]+1+n_write,
			   dep->Constraint[i][1+i], mone, dep->Dimension-n_write+1);
	    value_clear(mone);
	}
	if (value_notmone_p(dep->Constraint[i][1+n_write+i]))
	    return false;
	if (First_Non_Zero(dep->Constraint[i]+1+n_write, i) != -1)
	    return false;
	if (First_Non_Zero(dep->Constraint[i]+1+n_write+i+1, 
			   dep->Dimension-n_write-i-1-nparam) != -1)
	    return false;
    }
    for (int i = n_write; i < dep->NbConstraints; ++i) {
	int k = First_Non_Zero(dep->Constraint[i]+1, dep->Dimension-nparam);
	if (k == -1)
	    continue;
	int l = First_Non_Zero(dep->Constraint[i]+1+k+1, dep->Dimension-nparam-k-1);
	if (l == -1)
	    continue;
	if (value_notzero_p(dep->Constraint[i][0]))
	    return false;
	if (First_Non_Zero(dep->Constraint[i]+1+k+1+l+1, 
			   dep->Dimension-nparam-k-1-l-1) != -1)
	    return false;
    }

    return true;
}

static int selfloop_size_full(Polyhedron *dep, Polyhedron *reads,
			      int n_write, int n_read, Polyhedron *max,
			      int maxsize, unsigned MaxRays)
{
    Polyhedron *U = Universe_Polyhedron(0);
    evalue mone;
    value_init(mone.d);
    evalue_set_si(&mone, -1, 1);
    for (Polyhedron *read = reads; read; read = read->next) {
	Matrix *M = Matrix_Alloc(n_read, read->Dimension+2);
	for (int i = 0; i < n_read; ++i) {
	    value_set_si(M->p[i][1+i], -1);
	    value_oppose(M->p[i][1+read->Dimension], 
			 max->Constraint[i][1+max->Dimension]);
	}
	for (int i = n_read-1; i >= 0; --i) {
	    value_set_si(M->p[i][0], 1);
	    value_decrement(M->p[i][1+read->Dimension], M->p[i][1+read->Dimension]);
	    Polyhedron *I = AddConstraints(M->p[0], i+1, read, MaxRays);
	    evalue *EP = lexsmaller_domain(dep, I, n_write, U);
	    evalue *EC = lexsmaller_domain(reads, I, n_write, U);
	    Polyhedron_Free(I);
	    emul(&mone, EC);
	    eadd(EC, EP);
	    free_evalue_refs(EC);
	    free(EC);
	    reduce_evalue(EP);
	    int size = evalue_max(EP, MaxRays);
	    free_evalue_refs(EP);
	    free(EP);
	    if (size > maxsize)
		maxsize = size;
	}
	Matrix_Free(M);
    }
    free_evalue_refs(&mone);
    Polyhedron_Free(U);
    return maxsize;
}

void evalue_max_relation(evalue *e)
{
    if (value_notzero_p(e->d))
	return;
    if (e->x.p->type != relation)
	return;
    if (e->x.p->size > 2)
	return;
    enode *p = e->x.p;
    value_clear(e->d);
    *e = p->arr[1];
    free_evalue_refs(&(p->arr[0]));
    free(p);
}

evalue *evalue_max_domain(evalue *e)
{
    if (EVALUE_IS_ZERO(*e))
	return e;
    assert(value_zero_p(e->d));
    assert(e->x.p->type == partition);
    evalue_max_relation(&e->x.p->arr[1]);
    if (e->x.p->size == 2)
	return e;
    int i;
    for (i = e->x.p->size/2 - 1; i >= 1; --i) {
	evalue_max_relation(&e->x.p->arr[2*i+1]);
	if (eequal(&e->x.p->arr[1], &e->x.p->arr[2*i+1])) {
	    Domain_Free(EVALUE_DOMAIN(e->x.p->arr[2*i]));
	    free_evalue_refs(&e->x.p->arr[2*i+1]);
	    e->x.p->size -= 2;
	    continue;
	}
	if (value_zero_p(e->x.p->arr[2*i+1].d))
	    break;
	if (value_zero_p(e->x.p->arr[1].d))
	    break;
	assert(value_one_p(e->x.p->arr[2*i+1].d));
	assert(value_one_p(e->x.p->arr[1].d));
	if (value_gt(e->x.p->arr[2*i+1].x.n, e->x.p->arr[1].x.n)) {
	    Domain_Free(EVALUE_DOMAIN(e->x.p->arr[0]));
	    free_evalue_refs(&e->x.p->arr[1]);
	    e->x.p->arr[0] = e->x.p->arr[2*i];
	    e->x.p->arr[1] = e->x.p->arr[2*i+1];
	} else {
	    Domain_Free(EVALUE_DOMAIN(e->x.p->arr[2*i]));
	    free_evalue_refs(&e->x.p->arr[2*i+1]);
	}
	e->x.p->size -= 2;
    }
    if (i >= 1) {
	free_evalue_refs(e);
	free(e);
	return NULL;
    }
    return e;
}

evalue *evalue_max(evalue *e1, evalue *e2)
{
    if (!e1)
	return e2;
    assert(e2);
    if (EVALUE_IS_ZERO(*e1)) {
	free_evalue_refs(e1);
	free(e1);
	return e2;
    }
    if (EVALUE_IS_ZERO(*e2)) {
	free_evalue_refs(e2);
	free(e2);
	return e1;
    }
    assert(value_zero_p(e1->d));
    assert(e1->x.p->type == partition);
    assert(e1->x.p->size == 2);
    assert(value_zero_p(e2->d));
    assert(e2->x.p->type == partition);
    assert(e2->x.p->size == 2);
    if (!universeQ(EVALUE_DOMAIN(e1->x.p->arr[0])) ||
	    !universeQ(EVALUE_DOMAIN(e2->x.p->arr[0])) ||
	    !eequal(&e1->x.p->arr[1], &e2->x.p->arr[1])) {
	free_evalue_refs(e1);
	free(e1);
	free_evalue_refs(e2);
	free(e2);
	return NULL;
    }
    free_evalue_refs(e2);
    free(e2);
    return e1;
}

size::enumerator *selfloop_size_param(pdg::PDG *pdg, Polyhedron *dep,
				      Polyhedron *reads,
				      int dim, Polyhedron *max, Polyhedron *ctx, 
				      bool rect, size_options *options)
{
    if (!rect)
	return selfloop_size_bernstein(pdg, dep, reads, dim, max, ctx, options);

    size::enumerator *res = NULL;
    evalue mone;
    value_init(mone.d);
    evalue_set_si(&mone, -1, 1);

    assert(max);
    evalue *EP = lexsmaller_domain(dep, max, dim, ctx);
    evalue *EC = lexsmaller_domain(reads, max, dim, ctx);
    emul(&mone, EC);
    eadd(EC, EP);
    free_evalue_refs(EC);
    free(EC);
    reduce_evalue(EP);

    evalue_frac2floor(EP);
    evalue_remove_iterators(EP, max, dim, ctx);

    res = new enumerator(EP, pdg);

    free_evalue_refs(&mone);

    return res;
}

Matrix *set_M_from_max(Polyhedron *P_max, int n_write, int nparam)
{
    Matrix *M = Matrix_Alloc(n_write, n_write+nparam+2);
    for (int i = 0; i < n_write; ++i) {
	value_set_si(M->p[i][1+i], 1);
	for (int j = 0; j <= nparam; ++j)
	    value_assign(M->p[i][1+n_write+j], 
			 P_max->Constraint[i][1+P_max->Dimension-nparam+j]);
    }
    return M;
}

struct channel_counter : public clast_scanner {
    char tmpfile[L_tmpnam];
    char buffer[L_tmpnam+40];

    int count();
    void print_user(clast_user_stmt *s);
    void print_full(CloogNames* names, clast_stmt *s);
};

void channel_counter::print_full(CloogNames* names, clast_stmt *s)
{
    int ret;
    tmpnam(tmpfile);
    snprintf(buffer, 200, "gcc -x c -O2 -shared - -o %s", tmpfile);
    dst = popen(buffer, "w");
    assert(dst);
    print_header();
    fprintf(dst, "long __size_count() {\n");
    indent += 4;
    print_iterators(names);
    print_indent();
    fprintf(dst, "long count = 0, max = 0;\n");

    print(s);

    print_indent();
    fprintf(dst, "assert(count == 0);\n");
    fprintf(dst, "return max;\n");
    fprintf(dst, "}\n");
    ret = pclose(dst);
    assert(WEXITSTATUS(ret) == 0);
}

int channel_counter::count() {
    long (*c)();
    void *module;
    long max;
    char *error;

    module = dlopen(tmpfile, RTLD_LAZY);
    unlink(tmpfile);
    if ((error = dlerror()) != NULL) {
	fprintf(stderr, "%s\n", error);
	assert(0);
    }
    assert(module);
    c = (long (*)()) dlsym(module, "__size_count");
    if ((error = dlerror()) != NULL) {
	fprintf(stderr, "%s\n", error);
	assert(0);
    }
    max = c();
    dlclose(module);
    return max;
}

void channel_counter::print_user(clast_user_stmt *u)
{
    print_indent();
    if (u->statement->number == 1) {
	fprintf(dst, "if (++count > max)\n");
	print_indent();
	fprintf(dst, "    max = count;\n");
    } else
	fprintf(dst, "--count;\n");
}

static int scan_program(CloogProgram * p, CloogOptions *options)
{
    int max;
    clast_stmt *stmt = cloog_util_create_clast(p, options);

    channel_counter cs;
    cs.print_full(p->names, stmt);
    cloog_clast_free(stmt);
    max = cs.count();

    return max;
}

static int selfloop_size_fixed(Polyhedron *dep, Polyhedron *reads)
{
    CloogOptions *options = cloog_options_malloc();
    options->compilable = 1;

    cloog_program_builder pb;

    pb.p->nb_scattdims = 0;
    pb.p->scaldims = NULL;

    for (Polyhedron *P = reads; P; P = P->next) {
	pb.add_domain(Polyhedron_Copy(P), -1);
    }
    for (Polyhedron *P = dep; P; P = P->next) {
	pb.add_domain(Polyhedron_Copy(P), 1);
    }

    seq<pdg::parameter> params;
    pb.set_names(params, 0);

    //cloog_program_dump_cloog(stderr, pb.p);
    pb.p = cloog_program_generate(pb.p, options);
    int size = scan_program(pb.p, options);
    cloog_program_free(pb.p);
    cloog_options_free(options);
    return size;
}

enumerator::operator pdg::enumerator *() const
{
    switch (type) {
    case fixed:
	return new pdg::enumerator(v);
    case polylib:
	assert(!EVALUE_IS_ZERO(*EP));
	return new pdg::enumerator(new string(evalue2string(EP, pdg->params)));
#ifdef HAVE_BERNSTEIN
    case ginac:
	std::ostringstream strm;
	strm << *pl;
	return new pdg::enumerator(new string(strm.str()));
#endif
    }
}

integer *enumerator::evaluate() const
{
    switch (type) {
    case fixed:
	return new integer(v);
    case polylib: {
	Value params[pdg->params.size()];
	for (int i = 0; i < pdg->params.size(); ++i)
	    if (!pdg->params[i]->value)
		return NULL;
	for (int i = 0; i < pdg->params.size(); ++i) {
	    value_init(params[i]);
	    value_set_si(params[i], pdg->params[i]->value->v);
	}
	int v = (int) compute_evalue(EP, params);
	for (int i = 0; i < pdg->params.size(); ++i)
	    value_clear(params[i]);
	return new integer(v);
    }
#ifdef HAVE_BERNSTEIN
    case ginac: {
	GiNaC::exvector values;
	for (int i = 0; i < pdg->params.size(); ++i)
	    values.push_back(pdg->params[i]->value->v);
	GiNaC::numeric v = pl->evaluate(values);
	v = cln::floor1(cln::the<cln::cl_RA>(v.to_cl_N()));
	return new integer(v.to_int());
    }
#endif
    }
    return NULL;
}

size::enumerator *selfloop_size(pdg::PDG *pdg, Polyhedron *dep,
			       int n_write, int n_read, Polyhedron *ctx,
			       size_options *options)
{
    int nparam = ctx->Dimension;
    int extra = dep->Dimension - n_write - n_read - nparam;
    Polyhedron *reads = NULL;
    Polyhedron *max = NULL;

    bool rect = is_rectangular_dependence(dep, n_write, n_read, nparam);
    assert(n_write == n_read);

    Matrix *M = NULL;
    for (Polyhedron *P = dep; P; P = P->next) {
	Polyhedron *read = Polyhedron_projectout_dep(P, n_write);
	reads = Domain_concat(read, reads);
	if (nparam && !rect)
	    continue;
	Polyhedron *P_max = pip_lexminmax(P, 0, P->Dimension-nparam, nparam, 1);
	if (P_max->Dimension > P->Dimension) {
	    /* The max depends on some new existential variables.
	     * set_M_from_max doesn't expect this situation.
	     * Let's just skip it for now.
	     */
	    Polyhedron_Free(P_max);
	    rect = false;
	    continue;
	}
	if (!M) {
	    M = set_M_from_max(P_max, n_write, nparam);
	} else {
	    int i;
	    for (i = 0; i < n_write; ++i)
		if (value_lt(P_max->Constraint[i][1+P->Dimension], 
			     M->p[i][1+n_write+nparam]))
		    break;
	    if (i < n_write)
		for (i = 0; i < n_write; ++i)
		    value_assign(M->p[i][1+n_write+nparam], 
				 P_max->Constraint[i][1+P->Dimension]);
	}
	Polyhedron_Free(P_max);
    }
    if ((nparam && rect) || (!nparam && !options->flags)) {
	value_decrement(M->p[n_write-1][1+n_write+nparam], 
			M->p[n_write-1][1+n_write+nparam]);
	max = Constraints2Polyhedron(M, options->barvinok->MaxRays);
    }
    if (M)
	Matrix_Free(M);

    size::enumerator *e = NULL;
    if (!nparam && options->flags != size_bernstein) {
	int maxsize;
	if (options->flags == size_simulate)
	    maxsize = selfloop_size_fixed(dep, reads);
	else {
	    int p = lexsmaller_point(dep, max, n_write);
	    int c = lexsmaller_point(reads, max, n_write);
	    maxsize = p-c;
	    if (p == -1 || c == -1)
		maxsize = -1;

	    if (!rect)
		maxsize = 
		    selfloop_size_full(dep, reads, n_write, n_read, max, maxsize,
				       options->barvinok->MaxRays);
	}
	e = new enumerator(maxsize);
    } else {
	e = selfloop_size_param(pdg, dep, reads, n_write, max, ctx, rect, options); 
    }

    Domain_Free(reads);
    Polyhedron_Free(max);
    return e;
}

int selfloop_size_old(Polyhedron *dep, int n_write, int n_read, Polyhedron *ctx,
		      unsigned MaxRays)
{
    int nparam = ctx->Dimension;
    int extra = dep->Dimension - n_write - n_read - nparam;
    int l = -1;
    Polyhedron *reads = NULL;
    Polyhedron *read_border = NULL;
    Polyhedron *write_border = NULL;

    if (nparam > 0)
	return -1;

    for (Polyhedron *P = dep; P; P = P->next) {
	Polyhedron *read = Polyhedron_projectout_dep(P, n_write);
	read->next = reads;
	reads = read;
	Polyhedron *read_left = pip_lexminmax(read, n_read-1, 1+extra, nparam, 0);
	read_border = Domain_concat(read_left, read_border);
	/* shift border to the right */
	value_decrement(P->Constraint[n_write-1][1+P->Dimension],
			P->Constraint[n_write-1][1+P->Dimension]);
	Polyhedron *write_right = pip_lexminmax(P, n_write-1, 1+n_read+extra, nparam, 1);
	value_increment(P->Constraint[n_write-1][1+P->Dimension],
			P->Constraint[n_write-1][1+P->Dimension]);
	write_border = Domain_concat(write_right, write_border);
    }
    int l1 = live(dep, reads, read_border, n_write, n_read, MaxRays);
    int l2 = l1 >= 0 ? live(dep, reads, write_border, n_write, n_read, MaxRays) : -1;
    if (l2 >= 0)
	l = l1 > l2 ? l1 : l2;
    Domain_Free(read_border);
    Domain_Free(write_border);
    Domain_Free(reads);

    return l;
}

}
