#include "srel.h"
#include "domain.h"
#include "pdgutil.h"

#include <vector>
#include <algorithm>

using pdg::varvector;
using pdg::max_index;
using pdg::constraint_vector;

SRelation::SRelation (int in, int out, std::vector< std::vector<int> > & c, 
    symtab_ann *a) : Relation(in, out), sa(a)
{
    assert(c.size() >= 1);
    int	dim = c[0].size();
    varvector vars;

    for (int j = 1; j <= n_inp(); ++j)
	vars.push_back(input_var(j));
    for (int j = 1; j <= n_out(); ++j)
	vars.push_back(output_var(j));
    assert(in + out + 2 == dim);

    F_And *base = add_and();
    for (int i = 0; i < c.size(); ++i) {
	Constraint_Handle h;
	if (c[i][0]) 
	    h = base->add_GEQ();
	else
	    h = base->add_EQ();
	for (int j = 1; j < dim-1; ++j)
	    h.update_coef(vars[j-1], c[i][j]);
	h.update_const(c[i][dim-1]);
    }
    finalize();
}

void SRelation::fix(var_sym *var, int val)
{
    sa->add_var(var);
    EQ_Handle e = and_with_EQ();
    e.update_coef(get_local(sa->vars[var]), -1);
    e.update_const(val);
}

/* Add constraint "var >= 0" */
void SRelation::set_infinite(var_sym *var)
{
    sa->add_var(var);
    GEQ_Handle e = and_with_GEQ();
    e.update_coef(get_local(sa->vars[var]), 1);
}

void SRelation::intersect(named_lin_ineq *nlq)
{
    for (int i = 1; i < nlq->n(); ++i)
	sa->add_var(nlq->names()[i].var());

    integer_matrix lq = nlq->ineqs();
    for (int i = 0; i < lq.m(); ++i) {
	GEQ_Handle g = and_with_GEQ();
	g.update_const(lq[i][0]);
	for (int j = 1; j < lq.n(); ++j) {
	    g.update_coef(get_local(sa->vars[nlq->names()[j].var()]), lq[i][j]);
	}
    }
    delete nlq;
}

static void print_constraint(Constraint_Handle c, varvector& vv, int geq)
{
    printf(" %6d", geq);
    for (int i = 0; i < vv.size(); ++i)
	printf(" %6d", c.get_coef(vv[i]));
    printf(" %6d", c.get_const());
    printf("\n");
}

void SRelation::dump()
{
    simplify();

    varvector vv;
    varvector params;
    if (is_set())
	for (int j = 1; j <= n_set(); ++j)
	    vv.push_back(set_var(j));
    else {
	for (int j = 1; j <= n_inp(); ++j)
	    vv.push_back(input_var(j));
	for (int j = 1; j <= n_out(); ++j)
	    vv.push_back(output_var(j));
    }

    const Variable_ID_Tuple * globals = global_decls();
    for (int i = 0; i < globals->size(); ++i)
	params.push_back(get_local((*globals)[i]));

    int d = 0;
    for (DNF_Iterator di(query_DNF()); di; ++di, ++d) {
	assert(d == 0);

	int c = 0;
	for (EQ_Iterator ei = (*di)->EQs(); ei; ++ei, ++c)
	    max_index((*ei), vv, params);
	for (GEQ_Iterator gi = (*di)->GEQs(); gi; ++gi, ++c)
	    max_index((*gi), vv, params);
	for (int i = 0; i < params.size(); ++i)
	    vv.push_back(params[i]);

	printf("%d %d\n", c, vv.size() + 2);
	for (EQ_Iterator ei = (*di)->EQs(); ei; ++ei)
	    print_constraint((*ei), vv, 0);
	for (GEQ_Iterator gi = (*di)->GEQs(); gi; ++gi)
	    print_constraint((*gi), vv, 1);
    }
}

std::vector<std::vector<std::vector<int> > > SRelation::constraints()
{
    std::vector<std::vector<std::vector<int> > > u;
    simplify();

    varvector vv;
    varvector params;
    if (is_set())
	for (int j = 1; j <= n_set(); ++j)
	    vv.push_back(set_var(j));
    else {
	for (int j = 1; j <= n_inp(); ++j)
	    vv.push_back(input_var(j));
	for (int j = 1; j <= n_out(); ++j)
	    vv.push_back(output_var(j));
    }

    const Variable_ID_Tuple * globals = global_decls();
    for (int i = 0; i < globals->size(); ++i)
	params.push_back(get_local((*globals)[i]));

    int d = 0;
    for (DNF_Iterator di(query_DNF()); di; ++di, ++d) {
	std::vector<std::vector<int> > poly;
	assert(d == 0);

	int c = 0;
	for (EQ_Iterator ei = (*di)->EQs(); ei; ++ei, ++c)
	    max_index((*ei), vv, params);
	for (GEQ_Iterator gi = (*di)->GEQs(); gi; ++gi, ++c)
	    max_index((*gi), vv, params);
	for (int i = 0; i < params.size(); ++i)
	    vv.push_back(params[i]);

	for (EQ_Iterator ei = (*di)->EQs(); ei; ++ei)
	    poly.push_back(constraint_vector((*ei), vv, 0));
	for (GEQ_Iterator gi = (*di)->GEQs(); gi; ++gi)
	    poly.push_back(constraint_vector((*gi), vv, 1));
	u.push_back(poly);
    }
    return u;
}

SRelation * SRelation::Intersection(SRelation *b)
{
    Relation I = ::Intersection(copy(*(Relation*) this), copy(*(Relation*) b));
    return new SRelation(I, sa);
}

SRelation * SRelation::Union(SRelation *b)
{
    Relation I = ::Union(copy(*(Relation*) this), copy(*(Relation*) b));
    return new SRelation(I, sa);
}

SRelation * SRelation::Difference(SRelation *b)
{
    Relation I = ::Difference(copy(*(Relation*) this), copy(*(Relation*) b));
    return new SRelation(I, sa);
}

SRelation * SRelation::Composition(SRelation *b)
{
    Relation I = ::Composition(copy(*(Relation*) this), copy(*(Relation*) b));
    return new SRelation(I, sa);
}

SRelation * SRelation::Complement()
{
    Relation I = ::Complement(copy(*(Relation*) this));
    return new SRelation(I, sa);
}

SMap::SMap (int in, int out, named_lin_ineq *nli, 
	    std::vector<var_sym *>& globals) : UnionMap(in, out)
{
    int nparams = globals.size();
    for (int i = 0; i < nparams; ++i) {
	int pos = nli->names().find(globals[i]);
	if (pos == -1)
	    nli->add_col(name_table_entry(globals[i]), 1 + in + out + i);
	else if (pos != 1 + in + out + i)
	    nli->swap_col(pos, 1 + in + out +i);
    }

    integer_matrix lq = nli->ineqs();

    /* Move constants to the end and add column of 1s in front */
    constraints.push_back(new pdg::Matrix());
    std::vector< std::vector<int> >& el = constraints[0]->el;
    for (int i = 0; i < lq.m(); ++i) {
	el.push_back(vector<int>(1+lq.n()));
	el[i][0] = 1;
	el[i][lq.n()] = lq[i][0];
	for (int j = 1; j < lq.n(); ++j)
	    el[i][j] = lq[i][j];
    }
    /* Add dummy constraint */
    if (el.size() == 0) {
	el.push_back(vector<int>(1+in+out+nparams+1));
	el[0][0] = 1;
	el[0][1+in+out+nparams] = 1;
    }
}

static at_init register_smap(SMap::register_type);

void SMap::register_type() {
    pdg::UnionMap::register_type();
    t_desc()[&typeid(SMap)] = t_desc()[&typeid(pdg::UnionMap)];
}
